package java.sql;

import java.util.NoSuchElementException;
import java.util.Iterator;
import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;

public class SQLException extends Exception implements Iterable<Throwable>
{
    private String SQLState;
    private int vendorCode;
    private volatile SQLException next;
    private static final AtomicReferenceFieldUpdater<SQLException, SQLException> nextUpdater;
    private static final long serialVersionUID = 2135244094396331484L;
    
    public SQLException(String message, String s, int n) {
        super(message);
        this.SQLState = s;
        this.vendorCode = n;
        if (!(this instanceof SQLWarning) && DriverManager.getLogWriter() != null) {
            DriverManager.println("SQLState(" + s + ") vendor code(" + n + ")");
            this.printStackTrace(DriverManager.getLogWriter());
        }
    }
    
    public SQLException(String message, String s) {
        super(message);
        this.SQLState = s;
        this.vendorCode = 0;
        if (!(this instanceof SQLWarning) && DriverManager.getLogWriter() != null) {
            this.printStackTrace(DriverManager.getLogWriter());
            DriverManager.println("SQLException: SQLState(" + s + ")");
        }
    }
    
    public SQLException(String message) {
        super(message);
        this.SQLState = null;
        this.vendorCode = 0;
        if (!(this instanceof SQLWarning) && DriverManager.getLogWriter() != null) {
            this.printStackTrace(DriverManager.getLogWriter());
        }
    }
    
    public SQLException() {
        this.SQLState = null;
        this.vendorCode = 0;
        if (!(this instanceof SQLWarning) && DriverManager.getLogWriter() != null) {
            this.printStackTrace(DriverManager.getLogWriter());
        }
    }
    
    public SQLException(Throwable cause) {
        super(cause);
        if (!(this instanceof SQLWarning) && DriverManager.getLogWriter() != null) {
            this.printStackTrace(DriverManager.getLogWriter());
        }
    }
    
    public SQLException(String message, Throwable cause) {
        super(message, cause);
        if (!(this instanceof SQLWarning) && DriverManager.getLogWriter() != null) {
            this.printStackTrace(DriverManager.getLogWriter());
        }
    }
    
    public SQLException(String message, String sqlState, Throwable cause) {
        super(message, cause);
        this.SQLState = sqlState;
        this.vendorCode = 0;
        if (!(this instanceof SQLWarning) && DriverManager.getLogWriter() != null) {
            this.printStackTrace(DriverManager.getLogWriter());
            DriverManager.println("SQLState(" + this.SQLState + ")");
        }
    }
    
    public SQLException(String message, String sqlState, int n, Throwable cause) {
        super(message, cause);
        this.SQLState = sqlState;
        this.vendorCode = n;
        if (!(this instanceof SQLWarning) && DriverManager.getLogWriter() != null) {
            DriverManager.println("SQLState(" + this.SQLState + ") vendor code(" + n + ")");
            this.printStackTrace(DriverManager.getLogWriter());
        }
    }
    
    public String getSQLState() {
        return this.SQLState;
    }
    
    public int getErrorCode() {
        return this.vendorCode;
    }
    
    public SQLException getNextException() {
        return this.next;
    }
    
    public void setNextException(SQLException ex) {
        while (true) {
            SQLException next = this.next;
            if (next != null) {
                this = next;
            }
            else {
                if (SQLException.nextUpdater.compareAndSet(this, null, ex)) {
                    break;
                }
                this = this.next;
            }
        }
    }
    
    @Override
    public Iterator<Throwable> iterator() {
        return new Iterator<Throwable>() {
            SQLException firstException = SQLException.this;
            SQLException nextException = this.firstException.getNextException();
            Throwable cause = this.firstException.getCause();
            
            @Override
            public boolean hasNext() {
                return this.firstException != null || this.nextException != null || this.cause != null;
            }
            
            @Override
            public Throwable next() {
                SQLException ex;
                if (this.firstException != null) {
                    ex = this.firstException;
                    this.firstException = null;
                }
                else if (this.cause != null) {
                    ex = this.cause;
                    this.cause = this.cause.getCause();
                }
                else if (this.nextException != null) {
                    ex = this.nextException;
                    this.cause = this.nextException.getCause();
                    this.nextException = this.nextException.getNextException();
                }
                else {
                    throw new NoSuchElementException();
                }
                return ex;
            }
            
            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }
    
    static {
        nextUpdater = AtomicReferenceFieldUpdater.newUpdater(SQLException.class, SQLException.class, "next");
    }
}
